require './exceptions'

def convert_syntax(val)
  # Convert rgbds syntax [] to non-rgbds syntax () if needed
  if val.is_a?(String)
    val.tr('[]','()')
  else
    val
  end
end

# May raise: ArgumentError for things that are not numbers but prefixed as such
def convert_numbers(val)
  # rgbds syntax:
  #   - decimal
  # % - binary
  # & - octal
  # $ - hexadecimal
  return 0 if val == '0'

  case val[0]
  when '%'
    Integer(val[1..-1], 2)
  when '&'
    Integer(val[1..-1], 8)
  when '$'
    Integer(val[1..-1], 16)
  else
    val2 = val.to_i
    if val2 == 0
      val
    else
      val2
    end
  end
end

def lex_line(line)
  if !line || line.empty? || line =~ /^\s*;/
    return nil
  end

  # Filter mid-line comments
  line = line.sub(/\s*;.*$/, '').strip

  inst, args = line.split(/\s+/, 2)

  # Convert labels to symbols
  inst = inst[1..-1].to_sym if inst[0] == '.'

  # no args: e.g. nop
  # 1 arg: e.g. inc d
  # 2 args: ld a,$ff

  args = if args
           args.split(/,\s*/)
         else
           []
         end

  args.map! do |arg|
    begin
      # Convert labels to symbols
      if arg[0] == '.'
        arg[1..-1].to_sym
      else
        convert_numbers(arg)
      end
    rescue ArgumentError
      raise LexException.new("Not a number but prefixed: #{arg}")
    end
  end

  args.map! {|arg| convert_syntax(arg)}

  [inst, args]
end

def lex(input)
  lexed = []
  errors = []

  input.each_line.with_index do |line, num|
    begin
      line.strip!
      line.downcase!
      res = lex_line(line)
      lexed << res if res
    rescue StandardError => e
      errors << ("Lexer error on line #{num}: #{e.message}\n" + e.backtrace.join("\n"))
    end
  end

  if !errors.empty?
    puts errors.join("\n")
    exit(1)
  end

  lexed
end

#p lex("ld a,$ff\nnop\ninc b")
#[["ld", ["a", 255]], ["nop", []], ["inc", ["b"]]]

